@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# © Copyright EnterpriseDB UK Limited 2011-2022
+# © Copyright EnterpriseDB UK Limited 2011-2023
 #
 # This file is part of Barman.
 #
@@ -37,7 +37,7 @@ import barman.config
 import barman.diagnose
 from barman import output
 from barman.annotations import KeepManager
-from barman.config import RecoveryOptions
+from barman.config import RecoveryOptions, parse_recovery_staging_path
 from barman.exceptions import (
     BadXlogSegmentName,
     RecoveryException,
@@ -48,6 +48,7 @@ from barman.infofile import BackupInfo, WalFileInfo
 from barman.server import Server
 from barman.utils import (
     BarmanEncoder,
+    check_backup_name,
     check_non_negative,
     check_positive,
     check_tli,
@@ -55,7 +56,9 @@ from barman.utils import (
     drop_privileges,
     force_str,
     get_log_levels,
+    get_backup_id_using_shortcut,
     parse_log_level,
+    RESERVED_BACKUP_IDS,
     SHA256,
 )
 from barman.xlog import check_archive_usable
@@ -334,7 +337,7 @@ def backup_completer(prefix, parsed_args, **kwargs):
     for backup_id in sorted(backups, reverse=True):
         if backup_id.startswith(prefix):
             yield backup_id
-    for special_id in ("latest", "last", "oldest", "first", "last-failed"):
+    for special_id in RESERVED_BACKUP_IDS:
         if len(backups) > 0 and special_id.startswith(prefix):
             yield special_id
 
@@ -395,6 +398,18 @@ def backup_completer(prefix, parsed_args, **kwargs):
             type=check_positive,
             metavar="NJOBS",
         ),
+        argument(
+            "--jobs-start-batch-period",
+            help="The time period in seconds over which a single batch of jobs will "
+            "be started.",
+            type=check_positive,
+        ),
+        argument(
+            "--jobs-start-batch-size",
+            help="The maximum number of parallel Rsync jobs to start in a single "
+            "batch.",
+            type=check_positive,
+        ),
         argument(
             "--bwlimit",
             help="maximum transfer rate in kilobytes per second. "
@@ -421,6 +436,30 @@ def backup_completer(prefix, parsed_args, **kwargs):
             default=None,
             type=check_non_negative,
         ),
+        argument(
+            "--name",
+            help="a name which can be used to reference this backup in barman "
+            "commands such as recover and delete",
+            dest="backup_name",
+            default=None,
+            type=check_backup_name,
+        ),
+        argument(
+            "--manifest",
+            help="forces the creation of the backup manifest file for the "
+            "rsync backup method",
+            dest="automatic_manifest",
+            action="store_true",
+            default=SUPPRESS,
+        ),
+        argument(
+            "--no-manifest",
+            help="disables the creation of the backup manifest file for the "
+            "rsync backup method",
+            dest="automatic_manifest",
+            action="store_false",
+            default=SUPPRESS,
+        ),
     ]
 )
 def backup(args):
@@ -442,13 +481,32 @@ def backup(args):
         if args.retry_times is not None:
             server.config.basebackup_retry_times = args.retry_times
         if hasattr(args, "immediate_checkpoint"):
+            # As well as overriding the immediate_checkpoint value in the config
+            # we must also update the immediate_checkpoint attribute on the
+            # postgres connection because it has already been set from the config
             server.config.immediate_checkpoint = args.immediate_checkpoint
+            server.postgres.immediate_checkpoint = args.immediate_checkpoint
+        if hasattr(args, "automatic_manifest"):
+            # Override the set value for the autogenerate_manifest config option.
+            # The backup executor class will automatically ignore --manifest requests
+            # for backup methods different from rsync.
+            server.config.autogenerate_manifest = args.automatic_manifest
         if args.jobs is not None:
             server.config.parallel_jobs = args.jobs
+        if args.jobs_start_batch_size is not None:
+            server.config.parallel_jobs_start_batch_size = args.jobs_start_batch_size
+        if args.jobs_start_batch_period is not None:
+            server.config.parallel_jobs_start_batch_period = (
+                args.jobs_start_batch_period
+            )
         if hasattr(args, "bwlimit"):
             server.config.bandwidth_limit = args.bwlimit
         with closing(server):
-            server.backup(wait=args.wait, wait_timeout=args.wait_timeout)
+            server.backup(
+                wait=args.wait,
+                wait_timeout=args.wait_timeout,
+                backup_name=args.backup_name,
+            )
     output.close_and_exit()
 
 
@@ -536,7 +594,7 @@ def replication_status(args):
     """
     Shows live information and status of any streaming client
     """
-    servers = get_server_list(args, skip_inactive=True)
+    servers = get_server_list(args, skip_inactive=True, skip_passive=True)
     for name in sorted(servers):
         server = servers[name]
 
@@ -663,6 +721,17 @@ def rebuild_xlogdb(args):
             type=check_positive,
             metavar="NJOBS",
         ),
+        argument(
+            "--jobs-start-batch-period",
+            help="The time period in seconds over which a single batch of jobs will "
+            "be started.",
+            type=check_positive,
+        ),
+        argument(
+            "--jobs-start-batch-size",
+            help="The maximum number of Rsync jobs to start in a single batch.",
+            type=check_positive,
+        ),
         argument(
             "--get-wal",
             help="Enable the get-wal option during the recovery.",
@@ -709,6 +778,50 @@ def rebuild_xlogdb(args):
             default=SUPPRESS,
             help="Enable standby mode when starting the recovered PostgreSQL instance",
         ),
+        argument(
+            "--recovery-staging-path",
+            dest="recovery_staging_path",
+            help=(
+                "A path to a location on the recovery host where compressed backup "
+                "files will be staged during the recovery. This location must have "
+                "enough available space to temporarily hold the full compressed "
+                "backup. This option is *required* when recovering from a compressed "
+                "backup."
+            ),
+        ),
+        argument(
+            "--recovery-conf-filename",
+            dest="recovery_conf_filename",
+            help=(
+                "Name of the file to which recovery configuration options will be "
+                "added for PostgreSQL 12 and later (default: postgresql.auto.conf)."
+            ),
+        ),
+        argument(
+            "--snapshot-recovery-instance",
+            help="Instance where the disks recovered from the snapshots are attached",
+        ),
+        argument(
+            "--snapshot-recovery-zone",
+            help=(
+                "Zone containing the instance and disks for the snapshot recovery "
+                "(deprecated: replaced by --gcp-zone)"
+            ),
+        ),
+        argument(
+            "--gcp-zone",
+            help="Zone containing the instance and disks for the snapshot recovery",
+        ),
+        argument(
+            "--azure-resource-group",
+            help="Azure resource group containing the instance and disks for recovery "
+            "of a snapshot backup",
+        ),
+        argument(
+            "--aws-region",
+            help="The name of the AWS region containing the EC2 VM and storage "
+            "volumes for recovery of a snapshot backup",
+        ),
     ]
 )
 def recover(args):
@@ -728,6 +841,35 @@ def recover(args):
         )
         output.close_and_exit()
 
+    # If the backup to be recovered is compressed then there are additional
+    # checks to be carried out
+    if backup_id.compression is not None:
+        # Set the recovery staging path from the cli if it is set
+        if args.recovery_staging_path is not None:
+            try:
+                recovery_staging_path = parse_recovery_staging_path(
+                    args.recovery_staging_path
+                )
+            except ValueError as exc:
+                output.error("Cannot parse recovery staging path: %s", str(exc))
+                output.close_and_exit()
+            server.config.recovery_staging_path = recovery_staging_path
+        # If the backup is compressed but there is no recovery_staging_path
+        # then this is an error - the user *must* tell barman where recovery
+        # data can be staged.
+        if server.config.recovery_staging_path is None:
+            output.error(
+                "Cannot recover from backup '%s' of server '%s': "
+                "backup is compressed with %s compression but no recovery "
+                "staging path is provided. Either set recovery_staging_path "
+                "in the Barman config or use the --recovery-staging-path "
+                "argument.",
+                args.backup_id,
+                server.config.name,
+                backup_id.compression,
+            )
+            output.close_and_exit()
+
     # decode the tablespace relocation rules
     tablespaces = {}
     if args.tablespace:
@@ -780,6 +922,10 @@ def recover(args):
             server.config.recovery_options.remove(RecoveryOptions.GET_WAL)
     if args.jobs is not None:
         server.config.parallel_jobs = args.jobs
+    if args.jobs_start_batch_size is not None:
+        server.config.parallel_jobs_start_batch_size = args.jobs_start_batch_size
+    if args.jobs_start_batch_period is not None:
+        server.config.parallel_jobs_start_batch_period = args.jobs_start_batch_period
     if hasattr(args, "bwlimit"):
         server.config.bandwidth_limit = args.bwlimit
 
@@ -818,6 +964,57 @@ def recover(args):
             output.close_and_exit()
         server.config.network_compression = args.network_compression
 
+    if backup_id.snapshots_info is not None:
+        missing_args = []
+        if not args.snapshot_recovery_instance:
+            missing_args.append("--snapshot-recovery-instance")
+        if len(missing_args) > 0:
+            output.error(
+                "Backup %s is a snapshot backup and the following required arguments "
+                "have not been provided: %s",
+                backup_id.backup_id,
+                ", ".join(missing_args),
+            )
+            output.close_and_exit()
+        if tablespaces != {}:
+            output.error(
+                "Backup %s is a snapshot backup therefore tablespace relocation rules "
+                "cannot be used.",
+                backup_id.backup_id,
+            )
+            output.close_and_exit()
+        # Set the snapshot keyword arguments to be passed to the recovery executor
+        snapshot_kwargs = {
+            "recovery_instance": args.snapshot_recovery_instance,
+        }
+        # Special handling for deprecated snapshot_recovery_zone arg
+        if args.gcp_zone is None and args.snapshot_recovery_zone is not None:
+            args.gcp_zone = args.snapshot_recovery_zone
+        # Override provider-specific options in the config
+        for arg in (
+            "aws_region",
+            "azure_resource_group",
+            "gcp_zone",
+        ):
+            value = getattr(args, arg)
+            if value is not None:
+                setattr(server.config, arg, value)
+    else:
+        unexpected_args = []
+        if args.snapshot_recovery_instance:
+            unexpected_args.append("--snapshot-recovery-instance")
+        if len(unexpected_args) > 0:
+            output.error(
+                "Backup %s is not a snapshot backup but the following snapshot "
+                "arguments have been used: %s",
+                backup_id.backup_id,
+                ", ".join(unexpected_args),
+            )
+            output.close_and_exit()
+        # An empty dict is used so that snapshot-specific arguments are not passed to
+        # non-snapshot recovery executors
+        snapshot_kwargs = {}
+
     with closing(server):
         try:
             server.recover(
@@ -834,6 +1031,8 @@ def recover(args):
                 remote_command=args.remote_ssh_command,
                 target_action=getattr(args, "target_action", None),
                 standby_mode=getattr(args, "standby_mode", None),
+                recovery_conf_filename=args.recovery_conf_filename,
+                **snapshot_kwargs
             )
         except RecoveryException as exc:
             output.error(force_str(exc))
@@ -1376,7 +1575,10 @@ def receive_wal(args):
     The process uses the streaming protocol to receive WAL files
     from the PostgreSQL server.
     """
-    server = get_server(args)
+    should_skip_inactive = not (
+        args.create_slot or args.drop_slot or args.stop or args.reset
+    )
+    server = get_server(args, skip_inactive=should_skip_inactive)
     if args.stop and args.reset:
         output.error("--stop and --reset options are not compatible")
     # If the caller requested to shutdown the receive-wal process deliver the
@@ -1447,7 +1649,9 @@ def verify_backup(args):
     # Raises an error if wrong backup
     backup_info = parse_backup_id(server, args)
     # get backup path
-    output.info("Verifying backup %s on server %s" % (args.backup_id, args.server_name))
+    output.info(
+        "Verifying backup '%s' on server %s" % (args.backup_id, args.server_name)
+    )
 
     server.backup_manager.verify_backup(backup_info)
     output.close_and_exit()
@@ -1480,7 +1684,10 @@ def generate_manifest(args):
     )
     backup_manifest.create_backup_manifest()
 
-    output.info("Backup %s is valid on server %s" % (args.backup_id, args.server_name))
+    output.info(
+        "Backup manifest for backup '%s' successfully generated for server %s"
+        % (args.backup_id, args.server_name)
+    )
     output.close_and_exit()
 
 
@@ -1723,7 +1930,12 @@ def get_server(
     # Apply standard validation control and skips
     # the server if inactive or disabled, displaying standard
     # error messages. If on_error_stop (default) exits
-    if not manage_server_command(server, name, inactive_is_error) and on_error_stop:
+    if (
+        not manage_server_command(
+            server, name, inactive_is_error, skip_inactive=skip_inactive
+        )
+        and on_error_stop
+    ):
         output.close_and_exit()
         # The following return statement will never be reached
         # but it is here for clarity
@@ -1787,17 +1999,17 @@ def get_server_list(
         # When 'all' is used, it must be the only specified argument
         if args and len(args.server_name) != 1:
             output.error("You cannot use 'all' with other server names")
-        servers = available_servers
+        server_names = available_servers
     else:
         # Put servers in a set, so multiple occurrences are counted only once
-        servers = set(args.server_name)
+        server_names = set(args.server_name)
 
     # Loop through all the requested servers
-    for server in servers:
-        conf = barman.__config__.get_server(server)
+    for server_name in server_names:
+        conf = barman.__config__.get_server(server_name)
         if conf is None:
             # Unknown server
-            server_dict[server] = None
+            server_dict[server_name] = None
         else:
             server_object = Server(conf)
             # Skip inactive servers, if requested
@@ -1812,7 +2024,7 @@ def get_server_list(
             if skip_passive and server_object.passive_node:
                 output.info("Skipping passive server '%s'", conf.name)
                 continue
-            server_dict[server] = server_object
+            server_dict[server_name] = server_object
 
     return server_dict
 
@@ -1878,14 +2090,13 @@ def parse_backup_id(server, args):
     :param args: command line arguments namespace
     :rtype: barman.infofile.LocalBackupInfo
     """
-    if args.backup_id in ("latest", "last"):
-        backup_id = server.get_last_backup_id()
-    elif args.backup_id in ("oldest", "first"):
-        backup_id = server.get_first_backup_id()
-    elif args.backup_id in ("last-failed"):
-        backup_id = server.get_last_backup_id([BackupInfo.FAILED])
-    else:
-        backup_id = args.backup_id
+    backup_id = get_backup_id_using_shortcut(server, args.backup_id, BackupInfo)
+    if backup_id is None:
+        try:
+            backup_id = server.get_backup_id_from_name(args.backup_id)
+        except ValueError as exc:
+            output.error(str(exc))
+            output.close_and_exit()
     backup_info = server.get_backup(backup_id)
     if backup_info is None:
         output.error(
