@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# © Copyright EnterpriseDB UK Limited 2018-2022
+# © Copyright EnterpriseDB UK Limited 2018-2023
 #
 # This file is part of Barman.
 #
@@ -30,10 +30,13 @@ from barman.clients.cloud_cli import (
     OperationErrorExit,
 )
 from barman.cloud import CloudBackupCatalog, configure_logging
-from barman.cloud_providers import get_cloud_interface
-from barman.exceptions import InvalidRetentionPolicy
+from barman.cloud_providers import (
+    get_cloud_interface,
+    get_snapshot_interface_from_backup_info,
+)
+from barman.exceptions import BadXlogPrefix, InvalidRetentionPolicy
 from barman.retention_policies import RetentionPolicyFactory
-from barman.utils import force_str
+from barman.utils import check_non_negative, force_str
 from barman import xlog
 
 
@@ -102,6 +105,65 @@ def _remove_wals_for_backup(
             deleted_backup=deleted_backup,
             available_backups=catalog.get_backup_list(),
         )
+        # Identify any prefixes under which all WALs are no longer needed.
+        # This is a shortcut which allows us to delete all WALs under a prefix without
+        # checking each individual WAL.
+        try:
+            wal_prefixes = catalog.get_wal_prefixes()
+        except NotImplementedError:
+            # If fetching WAL prefixes isn't supported by the cloud provider then
+            # the old method of checking each WAL must be used for all WALs.
+            wal_prefixes = []
+        deletable_prefixes = []
+        for wal_prefix in wal_prefixes:
+            try:
+                tli_and_log = wal_prefix.split("/")[-2]
+                tli, log = xlog.decode_hash_dir(tli_and_log)
+            except (BadXlogPrefix, IndexError):
+                # If the prefix does not appear to be a tli and log we output a warning
+                # and move on to the next prefix rather than error out.
+                logging.warning(
+                    "Ignoring malformed WAL object prefix: {}".format(wal_prefix)
+                )
+                continue
+            # If this prefix contains a timeline which should be protected then we
+            # cannot delete the WALS under it so advance to the next prefix.
+            if tli in timelines_to_protect:
+                continue
+
+            # If the tli and log fall are inclusively between the tli and log for the
+            # begin and end WAL of any protected WAL range then this prefix cannot be
+            # deleted outright.
+            for begin_wal, end_wal in wal_ranges_to_protect:
+                begin_tli, begin_log, _ = xlog.decode_segment_name(begin_wal)
+                end_tli, end_log, _ = xlog.decode_segment_name(end_wal)
+                if (
+                    tli >= begin_tli
+                    and log >= begin_log
+                    and tli <= end_tli
+                    and log <= end_log
+                ):
+                    break
+            else:
+                # The prefix tli and log do not match any protected timelines or
+                # protected WAL ranges so all WALs are eligible for deletion if the tli
+                # is the same timeline and the log is below the begin_wal log of the
+                # backup being deleted.
+                until_begin_tli, until_begin_log, _ = xlog.decode_segment_name(
+                    remove_until.begin_wal
+                )
+                if tli == until_begin_tli and log < until_begin_log:
+                    # All WALs under this prefix pre-date the backup being deleted so they
+                    # can be deleted in one request.
+                    deletable_prefixes.append(wal_prefix)
+        for wal_prefix in deletable_prefixes:
+            if not dry_run:
+                cloud_interface.delete_under_prefix(wal_prefix)
+            else:
+                print(
+                    "Skipping deletion of all objects under prefix %s "
+                    "due to --dry-run option" % wal_prefix
+                )
         try:
             wal_paths = catalog.get_wal_paths()
         except Exception as exc:
@@ -112,6 +174,10 @@ def _remove_wals_for_backup(
             )
             return
         for wal_name, wal in wal_paths.items():
+            # If the wal starts with a prefix we deleted then ignore it so that the
+            # dry-run output is accurate
+            if any(wal.startswith(prefix) for prefix in deletable_prefixes):
+                continue
             if xlog.is_history_file(wal_name):
                 continue
             if timelines_to_protect:
@@ -163,19 +229,44 @@ def _delete_backup(
     cloud_interface,
     catalog,
     backup_id,
-    dry_run=True,
+    config,
     skip_wal_cleanup_if_standalone=True,
 ):
     backup_info = catalog.get_backup_info(backup_id)
     if not backup_info:
         logging.warning("Backup %s does not exist", backup_id)
         return
+
+    if backup_info.snapshots_info:
+        logging.debug(
+            "Will delete the following snapshots: %s",
+            ", ".join(
+                snapshot.identifier for snapshot in backup_info.snapshots_info.snapshots
+            ),
+        )
+        if not config.dry_run:
+            snapshot_interface = get_snapshot_interface_from_backup_info(
+                backup_info, config
+            )
+            snapshot_interface.delete_snapshot_backup(backup_info)
+        else:
+            print("Skipping deletion of snapshots due to --dry-run option")
+        # Delete the backup_label for snapshots backups as this is not stored in the
+        # same format used by the non-snapshot backups.
+        backup_label_path = os.path.join(
+            catalog.prefix, backup_info.backup_id, "backup_label"
+        )
+        if not config.dry_run:
+            cloud_interface.delete_objects([backup_label_path])
+        else:
+            print("Skipping deletion of %s due to --dry-run option" % backup_label_path)
+
     objects_to_delete = _get_files_for_backup(catalog, backup_info)
     backup_info_path = os.path.join(
         catalog.prefix, backup_info.backup_id, "backup.info"
     )
     logging.debug("Will delete backup.info file at %s" % backup_info_path)
-    if not dry_run:
+    if not config.dry_run:
         try:
             cloud_interface.delete_objects(objects_to_delete)
             # Do not try to delete backup.info until we have successfully deleted
@@ -192,7 +283,11 @@ def _delete_backup(
         )
 
     _remove_wals_for_backup(
-        cloud_interface, catalog, backup_info, dry_run, skip_wal_cleanup_if_standalone
+        cloud_interface,
+        catalog,
+        backup_info,
+        config.dry_run,
+        skip_wal_cleanup_if_standalone,
     )
     # It is important that the backup is removed from the catalog after cleaning
     # up the WALs because the code in _remove_wals_for_backup depends on the
@@ -238,23 +333,34 @@ def main(args=None):
                     % catalog.unreadable_backups
                 )
                 raise OperationErrorExit()
-
             if config.backup_id:
+                backup_id = catalog.parse_backup_id(config.backup_id)
                 # Because we only care about one backup, skip the annotation cache
                 # because it is only helpful when dealing with multiple backups
-                if catalog.should_keep_backup(config.backup_id, use_cache=False):
+                if catalog.should_keep_backup(backup_id, use_cache=False):
                     logging.error(
                         "Skipping delete of backup %s for server %s "
                         "as it has a current keep request. If you really "
                         "want to delete this backup please remove the keep "
                         "and try again.",
-                        config.backup_id,
+                        backup_id,
                         config.server_name,
                     )
                     raise OperationErrorExit()
-                _delete_backup(
-                    cloud_interface, catalog, config.backup_id, config.dry_run
-                )
+                if config.minimum_redundancy > 0:
+                    if config.minimum_redundancy >= len(catalog.get_backup_list()):
+                        logging.error(
+                            "Skipping delete of backup %s for server %s "
+                            "due to minimum redundancy requirements "
+                            "(minimum redundancy = %s, "
+                            "current redundancy = %s)",
+                            backup_id,
+                            config.server_name,
+                            config.minimum_redundancy,
+                            len(catalog.get_backup_list()),
+                        )
+                        raise OperationErrorExit()
+                _delete_backup(cloud_interface, catalog, backup_id, config)
             elif config.retention_policy:
                 try:
                     retention_policy = RetentionPolicyFactory.create(
@@ -262,6 +368,7 @@ def main(args=None):
                         config.retention_policy,
                         server_name=config.server_name,
                         catalog=catalog,
+                        minimum_redundancy=config.minimum_redundancy,
                     )
                 except InvalidRetentionPolicy as exc:
                     logging.error(
@@ -285,7 +392,7 @@ def main(args=None):
                         cloud_interface,
                         catalog,
                         backup_id,
-                        config.dry_run,
+                        config,
                         skip_wal_cleanup_if_standalone=False,
                     )
     except Exception as exc:
@@ -311,6 +418,13 @@ def parse_arguments(args=None):
         "--backup-id",
         help="Backup ID of the backup to be deleted",
     )
+    parser.add_argument(
+        "-m",
+        "--minimum-redundancy",
+        type=check_non_negative,
+        help="The minimum number of backups that should always be available.",
+        default=0,
+    )
     delete_arguments.add_argument(
         "-r",
         "--retention-policy",
@@ -323,6 +437,15 @@ def parse_arguments(args=None):
         action="store_true",
         help="Find the objects which need to be deleted but do not delete them",
     )
+    parser.add_argument(
+        "--batch-size",
+        dest="delete_batch_size",
+        type=int,
+        help="The maximum number of objects to be deleted in a single request to the "
+        "cloud provider. If unset then the maximum allowed batch size for the "
+        "specified cloud provider will be used (1000 for aws-s3, 256 for "
+        "azure-blob-storage and 100 for google-cloud-storage).",
+    )
     return parser.parse_args(args=args)
 
 
