@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# © Copyright EnterpriseDB UK Limited 2011-2022
+# © Copyright EnterpriseDB UK Limited 2011-2023
 #
 # This file is part of Barman.
 #
@@ -66,10 +66,13 @@ _SI_SUFFIX_RE = re.compile(r"""(\d+)\s*(k|Ki|M|Mi|G|Gi|T|Ti)?\s*$""")
 REUSE_BACKUP_VALUES = ("copy", "link", "off")
 
 # Possible copy methods for backups (must be all lowercase)
-BACKUP_METHOD_VALUES = ["rsync", "postgres", "local-rsync"]
+BACKUP_METHOD_VALUES = ["rsync", "postgres", "local-rsync", "snapshot"]
 
 CREATE_SLOT_VALUES = ["manual", "auto"]
 
+# Config values relating to pg_basebackup compression
+BASEBACKUP_COMPRESSIONS = ["gzip", "lz4", "zstd", "none"]
+
 
 class CsvOption(set):
 
@@ -287,6 +290,54 @@ def parse_reuse_backup(value):
     )
 
 
+def parse_backup_compression(value):
+    """
+    Parse a string to a valid backup_compression value.
+
+    :param str value: backup_compression value
+    :raises ValueError: if the value is invalid
+    """
+    if value is None:
+        return None
+    if value.lower() in BASEBACKUP_COMPRESSIONS:
+        return value.lower()
+    raise ValueError(
+        "Invalid value '%s'(must be one in: %s)" % (value, BASEBACKUP_COMPRESSIONS)
+    )
+
+
+def parse_backup_compression_format(value):
+    """
+    Parse a string to a valid backup_compression format value.
+
+    Valid values are "plain" and "tar"
+
+    :param str value: backup_compression_location value
+    :raises ValueError: if the value is invalid
+    """
+    if value is None:
+        return None
+    if value.lower() in ("plain", "tar"):
+        return value.lower()
+    raise ValueError("Invalid value (must be either `plain` or `tar`)")
+
+
+def parse_backup_compression_location(value):
+    """
+    Parse a string to a valid backup_compression location value.
+
+    Valid values are "client" and "server"
+
+    :param str value: backup_compression_location value
+    :raises ValueError: if the value is invalid
+    """
+    if value is None:
+        return None
+    if value.lower() in ("client", "server"):
+        return value.lower()
+    raise ValueError("Invalid value (must be either `client` or `server`)")
+
+
 def parse_backup_method(value):
     """
     Parse a string to a valid backup_method value.
@@ -305,6 +356,12 @@ def parse_backup_method(value):
     )
 
 
+def parse_recovery_staging_path(value):
+    if value is None or os.path.isabs(value):
+        return value
+    raise ValueError("Invalid value : '%s' (must be an absolute path)" % value)
+
+
 def parse_slot_name(value):
     """
     Replication slot names may only contain lower case letters, numbers,
@@ -325,6 +382,22 @@ def parse_slot_name(value):
     return value
 
 
+def parse_snapshot_disks(value):
+    """
+    Parse a comma separated list of names used to reference disks managed by a cloud
+    provider.
+
+    :param str value: Comma separated list of disk names
+    :return: List of disk names
+    """
+    disk_names = value.split(",")
+    # Verify each parsed disk is not an empty string
+    for disk_name in disk_names:
+        if disk_name == "":
+            raise ValueError(disk_names)
+    return disk_names
+
+
 def parse_create_slot(value):
     """
     Parse a string to a valid create_slot value.
@@ -354,6 +427,17 @@ class ServerConfig(object):
         "active",
         "archiver",
         "archiver_batch_size",
+        "autogenerate_manifest",
+        "aws_profile",
+        "aws_region",
+        "azure_credential",
+        "azure_resource_group",
+        "azure_subscription_id",
+        "backup_compression",
+        "backup_compression_format",
+        "backup_compression_level",
+        "backup_compression_location",
+        "backup_compression_workers",
         "backup_directory",
         "backup_method",
         "backup_options",
@@ -371,6 +455,8 @@ class ServerConfig(object):
         "disabled",
         "errors_directory",
         "forward_config_path",
+        "gcp_project",
+        "gcp_zone",
         "immediate_checkpoint",
         "incoming_wals_directory",
         "last_backup_maximum_age",
@@ -380,6 +466,8 @@ class ServerConfig(object):
         "minimum_redundancy",
         "network_compression",
         "parallel_jobs",
+        "parallel_jobs_start_batch_period",
+        "parallel_jobs_start_batch_size",
         "path_prefix",
         "post_archive_retry_script",
         "post_archive_script",
@@ -401,13 +489,21 @@ class ServerConfig(object):
         "pre_recovery_script",
         "pre_wal_delete_script",
         "pre_wal_delete_retry_script",
+        "primary_checkpoint_timeout",
+        "primary_conninfo",
         "primary_ssh_command",
         "recovery_options",
+        "recovery_staging_path",
         "create_slot",
         "retention_policy",
         "retention_policy_mode",
         "reuse_backup",
         "slot_name",
+        "snapshot_disks",
+        "snapshot_gcp_project",  # Deprecated, replaced by gcp_project
+        "snapshot_instance",
+        "snapshot_provider",
+        "snapshot_zone",  # Deprecated, replaced by gcp_zone
         "ssh_command",
         "streaming_archiver",
         "streaming_archiver_batch_size",
@@ -423,6 +519,17 @@ class ServerConfig(object):
     BARMAN_KEYS = [
         "archiver",
         "archiver_batch_size",
+        "autogenerate_manifest",
+        "aws_profile",
+        "aws_region",
+        "azure_credential",
+        "azure_resource_group",
+        "azure_subscription_id",
+        "backup_compression",
+        "backup_compression_format",
+        "backup_compression_level",
+        "backup_compression_location",
+        "backup_compression_workers",
         "backup_method",
         "backup_options",
         "bandwidth_limit",
@@ -435,6 +542,7 @@ class ServerConfig(object):
         "custom_decompression_filter",
         "custom_compression_magic",
         "forward_config_path",
+        "gcp_project",
         "immediate_checkpoint",
         "last_backup_maximum_age",
         "last_backup_minimum_size",
@@ -443,6 +551,8 @@ class ServerConfig(object):
         "minimum_redundancy",
         "network_compression",
         "parallel_jobs",
+        "parallel_jobs_start_batch_period",
+        "parallel_jobs_start_batch_size",
         "path_prefix",
         "post_archive_retry_script",
         "post_archive_script",
@@ -466,11 +576,14 @@ class ServerConfig(object):
         "pre_wal_delete_retry_script",
         "primary_ssh_command",
         "recovery_options",
+        "recovery_staging_path",
         "create_slot",
         "retention_policy",
         "retention_policy_mode",
         "reuse_backup",
         "slot_name",
+        "snapshot_gcp_project",  # Deprecated, replaced by gcp_project
+        "snapshot_provider",
         "streaming_archiver",
         "streaming_archiver_batch_size",
         "streaming_archiver_name",
@@ -483,6 +596,7 @@ class ServerConfig(object):
         "active": "true",
         "archiver": "off",
         "archiver_batch_size": "0",
+        "autogenerate_manifest": "false",
         "backup_directory": "%(barman_home)s/%(name)s",
         "backup_method": "rsync",
         "backup_options": "",
@@ -498,6 +612,9 @@ class ServerConfig(object):
         "minimum_redundancy": "0",
         "network_compression": "false",
         "parallel_jobs": "1",
+        "parallel_jobs_start_batch_period": "1",
+        "parallel_jobs_start_batch_size": "10",
+        "primary_checkpoint_timeout": "0",
         "recovery_options": "",
         "create_slot": "manual",
         "retention_policy_mode": "auto",
@@ -519,6 +636,12 @@ class ServerConfig(object):
         "active": parse_boolean,
         "archiver": parse_boolean,
         "archiver_batch_size": int,
+        "autogenerate_manifest": parse_boolean,
+        "backup_compression": parse_backup_compression,
+        "backup_compression_format": parse_backup_compression_format,
+        "backup_compression_level": int,
+        "backup_compression_location": parse_backup_compression_location,
+        "backup_compression_workers": int,
         "backup_method": parse_backup_method,
         "backup_options": BackupOptions,
         "basebackup_retry_sleep": int,
@@ -533,9 +656,14 @@ class ServerConfig(object):
         "max_incoming_wals_queue": int,
         "network_compression": parse_boolean,
         "parallel_jobs": int,
+        "parallel_jobs_start_batch_period": int,
+        "parallel_jobs_start_batch_size": int,
+        "primary_checkpoint_timeout": int,
         "recovery_options": RecoveryOptions,
+        "recovery_staging_path": parse_recovery_staging_path,
         "create_slot": parse_create_slot,
         "reuse_backup": parse_reuse_backup,
+        "snapshot_disks": parse_snapshot_disks,
         "streaming_archiver": parse_boolean,
         "streaming_archiver_batch_size": int,
         "slot_name": parse_slot_name,
@@ -650,6 +778,19 @@ class ServerConfig(object):
 
         return bwlimit
 
+    def update_msg_list_and_disable_server(self, msg_list):
+        """
+        Will take care of upgrading msg_list
+        :param msg_list: str|list can be either a string or a list of strings
+        """
+        if not msg_list:
+            return
+        if type(msg_list) is not list:
+            msg_list = [msg_list]
+
+        self.msg_list.extend(msg_list)
+        self.disabled = True
+
 
 class Config(object):
     """This class represents the barman configuration.
@@ -719,7 +860,7 @@ class Config(object):
             return None
         try:
             value = self._config.get(section, option, raw=False, vars=defaults)
-            if value.lower() == "none":
+            if value == "None":
                 value = none_value
             if value is not None:
                 value = self._QUOTE_RE.sub(lambda m: m.group(2), value)
