@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# © Copyright EnterpriseDB UK Limited 2011-2022
+# © Copyright EnterpriseDB UK Limited 2011-2023
 #
 # This file is part of Barman.
 #
@@ -36,6 +36,7 @@ import sys
 from argparse import ArgumentTypeError
 from abc import ABCMeta, abstractmethod
 from contextlib import contextmanager
+from dateutil import tz
 
 from distutils.version import Version
 
@@ -51,6 +52,8 @@ else:
     _text_type = unicode  # noqa
     _string_types = basestring  # noqa
 
+RESERVED_BACKUP_IDS = ("latest", "last", "oldest", "first", "last-failed")
+
 
 def drop_privileges(user):
     """
@@ -253,6 +256,35 @@ def total_seconds(timedelta):
         return (timedelta.microseconds + secs) / 10.0**6
 
 
+def timestamp(datetime_value):
+    """
+    Compatibility method because datetime.timestamp is not available in Python 2.7.
+
+    :param datetime.datetime datetime_value: A datetime object to be converted
+        into a timestamp.
+    :rtype: float
+    """
+    try:
+        return datetime_value.timestamp()
+    except AttributeError:
+        return total_seconds(
+            datetime_value - datetime.datetime(1970, 1, 1, tzinfo=tz.tzutc())
+        )
+
+
+def range_fun(*args, **kwargs):
+    """
+    Compatibility method required while we still support Python 2.7.
+
+    This can be removed when Python 2.7 support is dropped and calling code can
+    reference `range` directly.
+    """
+    try:
+        return xrange(*args, **kwargs)
+    except NameError:
+        return range(*args, **kwargs)
+
+
 def which(executable, path=None):
     """
     This method is useful to find if a executable is present into the
@@ -479,6 +511,7 @@ def with_metaclass(meta, *bases):
 
     :param type meta: Metaclass to add to base class
     """
+
     # This requires a bit of explanation: the basic idea is to make a
     # dummy metaclass for one level of class instantiation that replaces
     # itself with the actual metaclass.
@@ -495,6 +528,7 @@ def timeout(timeout_duration):
     ContextManager responsible for timing out the contained
     block of code after a defined time interval.
     """
+
     # Define the handler for the alarm signal
     def handler(signum, frame):
         raise TimeoutError()
@@ -745,3 +779,90 @@ def check_size(value):
     if int_value is None or int_value < 1:
         raise ArgumentTypeError("'%s' is not a valid size string" % value)
     return int_value
+
+
+def check_backup_name(backup_name):
+    """
+    Verify that a backup name is not a backup ID or reserved identifier.
+
+    Returns the backup name if it is a valid backup name and raises an exception
+    otherwise. A backup name is considered valid if it is not None, not empty,
+    does not match the backup ID format and is not any other reserved backup
+    identifier.
+
+    :param str backup_name: The backup name to be checked.
+    :return str: The backup name.
+    """
+    if backup_name is None:
+        raise ArgumentTypeError("Backup name cannot be None")
+    if backup_name == "":
+        raise ArgumentTypeError("Backup name cannot be empty")
+    if is_backup_id(backup_name):
+        raise ArgumentTypeError(
+            "Backup name '%s' is not allowed: backup ID" % backup_name
+        )
+    if backup_name in (RESERVED_BACKUP_IDS):
+        raise ArgumentTypeError(
+            "Backup name '%s' is not allowed: reserved word" % backup_name
+        )
+    return backup_name
+
+
+def is_backup_id(backup_id):
+    """
+    Checks whether the supplied identifier is a backup ID.
+
+    :param str backup_id: The backup identifier to check.
+    :return bool: True if the backup matches the backup ID regex, False otherwise.
+    """
+    return bool(re.match(r"(\d{8})T\d{6}$", backup_id))
+
+
+def get_backup_info_from_name(backups, backup_name):
+    """
+    Get the backup metadata for the named backup.
+
+    :param list[BackupInfo] backups: A list of BackupInfo objects which should be
+        searched for the named backup.
+    :param str backup_name: The name of the backup for which the backup metadata
+        should be retrieved.
+    :return BackupInfo|None: The backup metadata for the named backup.
+    """
+    matching_backups = [
+        backup for backup in backups if backup.backup_name == backup_name
+    ]
+    if len(matching_backups) > 1:
+        matching_backup_ids = " ".join(
+            [backup.backup_id for backup in matching_backups]
+        )
+        msg = (
+            "Multiple backups found matching name '%s' "
+            "(try using backup ID instead): %s"
+        ) % (backup_name, matching_backup_ids)
+
+        raise ValueError(msg)
+    elif len(matching_backups) == 1:
+        return matching_backups[0]
+
+
+def get_backup_id_using_shortcut(server, shortcut, BackupInfo):
+    """
+    Get backup ID from one of Barman shortcuts.
+
+    :param str server: The obj where to look from.
+    :param str shortcut: pattern to search.
+    :param BackupInfo BackupInfo: Place where we keep some Barman constants.
+    :return str backup_id|None: The backup ID for the provided shortcut.
+    """
+    backup_id = None
+
+    if shortcut in ("latest", "last"):
+        backup_id = server.get_last_backup_id()
+    elif shortcut in ("oldest", "first"):
+        backup_id = server.get_first_backup_id()
+    elif shortcut in ("last-failed"):
+        backup_id = server.get_last_backup_id([BackupInfo.FAILED])
+    elif is_backup_id(shortcut):
+        backup_id = shortcut
+
+    return backup_id
